/* discrepancy_search.c++

	Attempts to search for long low-discrepancy sequences
	Uses Angelo Furfaro's uQuadProg library (and indeed, this source
file is a hacked version of "example.c" which was distributed with
it.)
	Thus, this is distributed under the GPL, v.2 .

http://www.lis.deis.unical.it/~furfaro/uQuadProg/uQuadProg.php

Josh Burdick
jburdick@keyfitz.org

20100116 - initial version

 LICENSE
 
Copyright (2010) Josh Burdick

discrepancy_search.c++ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

discrepancy_search.c++ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with discrepancy_search.c++; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <iostream>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <boost/numeric/ublas/vector.hpp>
#include "uquadprog.hpp"
#include <iostream>

/** Utility for debugging. */
void printMatrix(matrix<double> &a) {
	for(unsigned int i=0; i<a.size1(); i++) {
		for(unsigned int j=0; j<a.size2(); j++)
			std::cout << a(i,j) << ' ';
		std::cout << std::endl;
	}
}

/** Looks for a sequence with low discrepancy.
	I think this could be phrased several ways; this way is sort of
	the "dual" way.
	Args:
		n - length of sequence to look for
	Side effects:
		writes the following to standard output:
		<n> <optimized objective function value>
		<x0>
		<x1>
		...
		<x_{n-1}> */
void lookForSequence(int n){

	// compute the matrix.  First, compute the matrix D, which gives
	// (for each spacing d) the numbers to be summed to compute the
	// discrepancy for that spacing d.
	matrix<double> D(n,n); 
	for(int d=1; d<=n; d++)
		for(int i=d; i<=n; i+=d)
			// Subscripts are a bit confusing: since the array indices are
			// 0-based, but the loops are 1-based.  So we subtract 1
			// from each array coordinate.
			D(d-1,i-1) = 1.0;

// for debugging
//	std::cerr << "D =" << std::endl;
//	printMatrix(D);	
//	std::cerr << std::endl;

	// the quadratic term of the objective function is the
	// squared sum of the discrepancies for each spacing, so
	// the matrix is D'D, where D is the above matrix encoding
	// the discrepancies.
	matrix<double> G = prod(trans(D), D);

	// set G = D' * D
//	G = prod(D, D);

	// the linear term of the objective function isn't used
	vector<double> g0(n);

// for debugging
//	std::cerr << "G =" << std::endl;
//	printMatrix(G);	
//	std::cerr << std::endl;

	// there's just one equality constraint.
	// This is the "kickstarting" constraint that x1 == 1.
	// ??? should this actually be (x1 >= 1) ?
	matrix<double> CE(n,3);
	vector<double> ce0(3);
	CE(0,0) = -1.0;
	ce0(0) = 1.0;
// you can add other constraints, such as "x4 = 1"
//	CE(3,1) = -1.0;
//	ce0(1) = 1.0;

	// there aren't any inequality constraints
	matrix<double> CI(n,0); 
	vector<double>ci0(0);

	// the vector we're optimizing
	vector<double>x(n);

	// optimize, and print results
	// FIXME: print discrepancy of solution?
	std::cout << n << ' ' << solve_quadprog(G, g0,  CE, ce0,  CI, ci0, x);
	std::cout << std::endl;
	for (unsigned int i = 0; i <x.size(); i++)
		std::cout << x(i) << ' ';
	std::cout << std::endl;

	// print the signs of the x's
	for (unsigned int j = 0; j <x.size(); j++)
		if (x(j) > 1e-10)
			std::cout << '+';
		else if (x(j) < -1e-10)
			std::cout << '-';
		else
			std::cout << '.';
	std::cout << std::endl;
}

int main(int argc, char** argv){
	if (argc != 2) {
		std::cerr << "discrepancy.c++" << std::endl;
		std::cerr << "looks for short +1/-1 sequences with low Erdos discrepancy" << std::endl;
		std::cerr << "Uses Angelo Furfaro's uQuadProg library" << std::endl;
		return 1;
	}

	int n = atoi(argv[1]);
	lookForSequence(n);
	return 0;
}

